/* ====================================================================
 * The Prothon License Agreement, Version 1.1
 *
 * Copyright (c) 2004 Hahn Creative Applications, http://hahnca.com.
 * All rights reserved. 
 *
 * 1. This LICENSE AGREEMENT is between Hahn Creative Applications ("HCA"),
 * and the Individual or Organization ("Licensee") accessing and otherwise
 * using Prothon software in source or binary form and its associated
 * documentation.
 * 
 * 2. Subject to the terms and conditions of this License Agreement, HCA
 * hereby grants Licensee a nonexclusive, royalty-free, world-wide license
 * to reproduce, analyze, test, perform and/or display publicly, prepare
 * derivative works, distribute, and otherwise use Prothon alone or in any
 * derivative version, provided, however, that HCA's License Agreement and
 * HCA's notice of copyright, i.e., "Copyright (c) 2004 Hahn Creative
 * Applications; All Rights Reserved" are retained in Prothon alone or
 * in any derivative version prepared by Licensee.
 * 
 * 3. In the event Licensee prepares a derivative work that is based on or
 * incorporates Prothon or any part thereof, and wants to make the
 * derivative work available to others as provided herein, then Licensee
 * hereby agrees to include in any such work a brief summary of the
 * changes made to Prothon.
 * 
 * 4. HCA is making Prothon available to Licensee on an "AS IS" basis.
 * HCA MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.  BY WAY
 * OF EXAMPLE, BUT NOT LIMITATION, HCA MAKES NO AND DISCLAIMS ANY
 * REPRESENTATION OR WARRANTY OF MERCHANTABILITY OR FITNESS FOR ANY
 * PARTICULAR PURPOSE OR THAT THE USE OF PROTHON WILL NOT INFRINGE ANY
 * THIRD PARTY RIGHTS.
 * 
 * 5. HCA SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PROTHON
 * FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL DAMAGES OR LOSS AS A
 * RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PROTHON, OR ANY
 * DERIVATIVE THEREOF, EVEN IF ADVISED OF THE POSSIBILITY THEREOF.
 * 
 * 6. This License Agreement will automatically terminate upon a material
 * breach of its terms and conditions.
 * 
 * 7. Nothing in this License Agreement shall be deemed to create any
 * relationship of agency, partnership, or joint venture between HCA and
 * Licensee.  This License Agreement does not grant permission to use HCA
 * trademarks or trade name in a trademark sense to endorse or promote
 * products or services of Licensee, or any third party.
 * 
 * 8. By copying, installing or otherwise using Prothon, Licensee agrees
 * to be bound by the terms and conditions of this License Agreement.
 * ====================================================================
 */


// File.c

#include <prothon/prothon_dll.h>
#include <stdio.h>
#include <string.h>

#include <apr_file_io.h>
#include <apr_strings.h>

#define MIN_BUFSIZE  65536

typedef struct {
	apr_file_t	*stream;	/* APR file reference */
	int			bufsize;	/* Size of our buffer */
	apr_pool_t	*pool;		/* Pool associated with this file */
	apr_int32_t	flags;		/* APR flags */
} pr_file_t;

typedef pr_file_t* pr_file_p;

#define STREAM(obj)  (((pr_file_p)((obj)->data.ptr))->stream)
#define BUFSIZE(obj) (((pr_file_p)((obj)->data.ptr))->bufsize)
#define POOL(obj)    (((pr_file_p)((obj)->data.ptr))->pool)
#define FLAGS(obj)   (((pr_file_p)((obj)->data.ptr))->flags)
#define OPEN(obj)    (((obj) != File_OBJ) && STREAM(obj))

MODULE_DECLARE(File);

static void set_file_obj(obj_p file_obj, obj_p filename, obj_p mode,
			     apr_file_t *fp, int bufsize,
			     apr_int32_t flags)
{
	apr_pool_t *subpool = apr_file_pool_get(fp);
	pr_file_p filep;

	set_attr(ist, file_obj, sym(ist, "name"), filename);
	set_attr(ist, file_obj, sym(ist, "mode"), mode);

	filep = pr_malloc(sizeof(*filep));

	if (bufsize < MIN_BUFSIZE)
		bufsize = MIN_BUFSIZE;

	filep->bufsize = bufsize;
	filep->stream  = fp;
	filep->pool = subpool;
	filep->flags = flags;
	SET_TYPE_IF_EXC(File_OBJ, file_obj, DATA_TYPE_DATAPTR) return;
	file_obj->data.ptr  = filep;
	file_obj->rd_access = ACC_USER1;
	file_obj->wr_access = ACC_USER1;
}


static obj_p new_file_object(obj_p filename, obj_p mode,
			     apr_file_t *fp, int bufsize,
			     apr_int32_t flags)
{
	obj_p file_obj;

	file_obj = NEW_OBJ(File_OBJ);
	set_file_obj(file_obj, filename, mode, fp, bufsize, flags);
	return file_obj;
}

MODULE_START(File)
{
	File_OBJ = NEW_OBJ(NULL);
	set_obj_doc(File_OBJ, "File object prototype");

	MODULE_ADD_TO_BASE(File);
	File_OBJ->unclonable = TRUE;
}

DEF(File, init_, FPARM3( path,    NULL, 
	                        mode,    NEW_STRING("r"),
                            bufsize, NEW_INT(-1)       ) ) {
	apr_file_t *f;
	apr_status_t aprerr;
	apr_pool_t *subpool;
	apr_int32_t flags = 0;
	char err_buf[1024];
	char *path, *mode;
	int bufsize;
	
	BIN_EMPTY_CHK();
	STRING_PARAM(1, path);
	STRING_PARAM(2, mode);
	INT_32_PARAM(3, bufsize);

	if ((mode[0] != 'r' && mode[0] != 'w' && mode[0] != 'a') ||
	    (mode[1] != '+' && mode[1] != '\0')) {
		raise_exception(ist, OBJ(TYPE_EXC), "File mode `%s' is not valid", mode);
		return NULL;
	}

	if (mode[0] == 'r') {
		flags = APR_READ;
		if (mode[1] == '+')
			flags |= APR_WRITE;
	} else if (mode[0] == 'w') {
		flags = APR_WRITE | APR_CREATE | APR_TRUNCATE;
		if (mode[1] == '+')
			flags |= APR_READ;
	} else if (mode[0] == 'a') {
		flags = APR_WRITE | APR_CREATE | APR_APPEND;
		if (mode[1] == '+')
			flags |= APR_READ;
	}

	aprerr = apr_pool_create(&subpool, get_pr_head_pool());
	if (aprerr != APR_SUCCESS) {
		raise_exception(ist, OBJ(OUTOFMEMORY_EXC), "Out of memory opening files");
		return NULL;
	}

	aprerr = apr_file_open(&f, path, flags, APR_OS_DEFAULT, subpool);
	if (aprerr != APR_SUCCESS) {
		raise_exception(ist, OBJ(IOEXCEPTION), "Unable to open path %s: %s",
				path, apr_strerror(aprerr, err_buf, sizeof(err_buf)));
		apr_pool_destroy(subpool);
		return NULL;
	}

	/* set_file_obj() modified the object, and thus calls write_lock.
	 * We need to unlock in order to avoid a deadlock */
	read_unlock(ist, self);
	set_file_obj(self, parms[1], parms[3], f, bufsize, flags);
	read_lock(ist, self);
	self->unclonable = TRUE;
	return OBJ(NONE);
}

DEF(File, objList_, FORM_RPARAM) {
	return parms[1];
}

DEF(File, close, NULL) {
	BIN_CONTENT_CHK(File);
	if (OPEN(self)) {
		apr_file_close(STREAM(self));
		STREAM(self) = NULL;
		apr_pool_destroy(POOL(self));
	}
	return OBJ(NONE);
}

DEF(File, closed_QUES, NULL) { 
	BIN_CONTENT_CHK(File);
	if (!has_proto(ist, self, File_OBJ)) {
		raise_exception(ist, OBJ(TYPE_EXC), "closed can only be called on a File object");
		return NULL;
	}
	if (OPEN(self)) return OBJ(PR_FALSE);
	else			return OBJ(PR_TRUE);
}

DEF(File, flush, NULL) { 
	BIN_CONTENT_CHK(File);
	if (!has_proto(ist, self, File_OBJ)) {
		raise_exception(ist, OBJ(TYPE_EXC), "flush can only be called on a File object");
		return NULL;
	}
	if (!OPEN(self)) {
		raise_exception(ist, OBJ(IOEXCEPTION), "flush called on closed file");
		return NULL;
	}
	if (apr_file_flush(STREAM(self)) == APR_SUCCESS) {
		raise_exception(ist, OBJ(IOEXCEPTION), "flushing file");
		return NULL;
	}
	return OBJ(NONE);
}
#if 0
DEF(File, fileNo, NULL) { 
	BIN_CONTENT_CHK(File);
	if (!has_proto(ist, self, File_OBJ)) {
		raise_exception(ist, OBJ(TYPE_EXC), "fileNo can only be called on a File object");
		return NULL;
	}
	if (!OPEN(self)) {
		raise_exception(ist, OBJ(IOEXCEPTION), "fileNo called on closed file");
		return NULL;
	}
	return NEW_INT(_fileno(STREAM(self)));
}

DEF(File, isATty_QUES, NULL) { 
	BIN_CONTENT_CHK(File);
	if (!OPEN(self)) {
		raise_exception(ist, OBJ(IOEXCEPTION), "isatty called on closed stream");
		return NULL;
	}
	if (_isatty(_fileno(STREAM(self))))  return OBJ(PR_TRUE); 
	else						return OBJ(PR_FALSE); 
}
#endif
DEF(File, read, FPARM1(size, NEW_INT(-1))) {
	int readall = FALSE;
	size_t size, total_read=0, num_read;
	obj_p str_obj;
	pr_str_p obj_str;
	char* buf;
	apr_status_t aprerr;
	char err_buf[1024];
	i64_t size_in;

	BIN_CONTENT_CHK(File);
	INT_64_PARAM(1, size_in);

	if (!has_proto(ist, self, File_OBJ)) {
		raise_exception(ist, OBJ(TYPE_EXC), "read can only be called on a File object");
		return NULL;
	}
	if (!OPEN(self)) {
		raise_exception(ist, OBJ(IOEXCEPTION), "read called on closed file");
		return NULL;
	}
	if (!(FLAGS(self) & APR_READ)) {
		raise_exception(ist, OBJ(IOEXCEPTION), "read called on file opened without read perms");
		return NULL;
	}
	if (!size_in) return NEW_STRING("");
	if (size_in < 0) {
		readall = TRUE;
		size = BUFSIZE(self);
	} else
		size = (size_t) size_in;
	str_obj = NEW_OBJ(OBJ(STRING_PROTO));
	if (size < IMMEDIATE_DATA_LEN) {
		SET_TYPE_IF_EXC(OBJ(STRING_PROTO), str_obj, DATA_TYPE_IMMDATA) return NULL;
		buf = str_obj->data.str;
	} else {
		obj_str = obj_malloc(ist, OBJ(STRING_PROTO), str_obj, sizeof(pr_str_t)+size+1);
		buf = obj_str->str;
	}
	num_read = size;
	aprerr = apr_file_read(STREAM(self), buf, &num_read);
	if (aprerr != APR_SUCCESS && aprerr != APR_EOF) {
		raise_exception(ist, OBJ(IOEXCEPTION), "%s",
				apr_strerror(aprerr, err_buf, sizeof(err_buf)));
		return NULL;
	}
	total_read += num_read;
	while(aprerr != APR_EOF && (readall || total_read < size)) {
		size_t next_read = (readall ? size : size - total_read);
		size_t new_len   =  total_read + next_read;
		if (str_obj->data_type == DATA_TYPE_IMMDATA && new_len >= IMMEDIATE_DATA_LEN) {
			obj_str = obj_malloc(ist, OBJ(STRING_PROTO), str_obj, sizeof(pr_str_t)+new_len+1);
			buf = obj_str->str;
		} else {
			obj_str = obj_realloc(ist, str_obj, sizeof(pr_str_t)+new_len+1);
			buf = obj_str->str;
		}
		num_read = next_read;
		aprerr = apr_file_read(STREAM(self), buf + total_read, &num_read);
		if (aprerr != APR_SUCCESS && aprerr != APR_EOF) {
			raise_exception(ist, OBJ(IOEXCEPTION), "%s",
					apr_strerror(aprerr, err_buf, sizeof(err_buf)));
			return NULL;
		}
		total_read += num_read;
	}
	if (str_obj->data_type == DATA_TYPE_IMMDATA)
		str_obj->imm_data_len = (int) total_read;
	else {
		obj_str = obj_realloc(ist, str_obj, sizeof(pr_str_t)+total_read+1);
		obj_str->len = total_read;
		buf = obj_str->str;
	}
	buf[total_read] = 0;
	str_obj->immutable = TRUE;
	return str_obj;
}

DEF(File, readLine, FPARM1(size, NEW_INT(-1))) {
	int readall = FALSE;
	size_t size, total_read;
	obj_p str_obj;
	pr_str_p obj_str;
	char* buf;
	apr_status_t aprerr;
	char err_buf[1024];
	i64_t  size_in;

	BIN_CONTENT_CHK(File);
	INT_64_PARAM(1, size_in);

	if (!has_proto(ist, self, File_OBJ)) {
		raise_exception(ist, OBJ(TYPE_EXC), "readLine can only be called on a File object");
		return NULL;
	}
	if (!OPEN(self)) {
		raise_exception(ist, OBJ(IOEXCEPTION), "readLine called on closed file");
		return NULL;
	}
	if (!(FLAGS(self) & APR_READ)) {
		raise_exception(ist, OBJ(IOEXCEPTION), "readLine called on file opened without read perms");
		return NULL;
	}
	if (!size_in) return NEW_STRING("");
	if (size_in < 0) {
		readall = TRUE;
		size = BUFSIZE(self);
	} else
		size = (size_t) size_in;
	str_obj = NEW_OBJ(OBJ(STRING_PROTO));
	if (size < IMMEDIATE_DATA_LEN) {
		SET_TYPE_IF_EXC(OBJ(STRING_PROTO), str_obj, DATA_TYPE_IMMDATA) return NULL;
		buf = str_obj->data.str;
	} else {
		obj_str = obj_malloc(ist, OBJ(STRING_PROTO), str_obj, sizeof(pr_str_t)+size+1);
		buf = obj_str->str;
	}
	aprerr = apr_file_gets(buf, (int)(size + 1), STREAM(self));
	if (aprerr != APR_SUCCESS && aprerr != APR_EOF) {
		raise_exception(ist, OBJ(IOEXCEPTION), "%s",
				apr_strerror(aprerr, err_buf, sizeof(err_buf)));
		return NULL;
	}
	total_read = strlen(buf);
	while (aprerr != APR_EOF && buf[total_read-1] != '\n' && (readall || total_read < size)) {
		size_t next_read = (readall ? size : size - total_read);
		size_t new_len = total_read + next_read;
		if (str_obj->data_type == DATA_TYPE_IMMDATA && new_len >= IMMEDIATE_DATA_LEN) {
			obj_str = obj_malloc(ist, OBJ(STRING_PROTO), str_obj, sizeof(pr_str_t)+new_len+1);
			buf = obj_str->str;
		} else {
			obj_str = obj_realloc(ist, str_obj, sizeof(pr_str_t)+new_len+1);
			buf = obj_str->str;
		}
		aprerr = apr_file_gets(buf + total_read, (int)(next_read + 1), STREAM(self));
		if (aprerr != APR_SUCCESS && aprerr != APR_EOF) {
			raise_exception(ist, OBJ(IOEXCEPTION), "%s",
					apr_strerror(aprerr, err_buf, sizeof(err_buf)));
			return NULL;
		}
		total_read = strlen(buf);
	}
	if (str_obj->data_type == DATA_TYPE_IMMDATA)
		str_obj->imm_data_len = (int) total_read;
	else {
		obj_str = obj_realloc(ist, str_obj, sizeof(pr_str_t)+total_read+1);
		buf = obj_str->str;
		obj_str->len = total_read;
	}
	buf[total_read] = 0;
	str_obj->immutable = TRUE;
	return str_obj;
}

DEF(File, readLines, FPARM1(size, NEW_INT(-1))) {
	int readall = FALSE;
	size_t size, total_read=0, num_read;
	obj_p str_obj, list_obj;
	pr_str_p obj_str;
	char* buf;
	apr_status_t aprerr;
	char err_buf[1024];
	i64_t  size_in;

	BIN_CONTENT_CHK(File);
	INT_64_PARAM(1, size_in);

	if (!has_proto(ist, self, File_OBJ)) {
		raise_exception(ist, OBJ(TYPE_EXC), "readLines can only be called on a File object");
		return NULL;
	}
	if (!OPEN(self)) {
		raise_exception(ist, OBJ(IOEXCEPTION), "readLines called on closed file");
		return NULL;
	}
        if (!(FLAGS(self) & APR_READ)) {
                raise_exception(ist, OBJ(IOEXCEPTION), "readLines called on file opened without read perms");
                return NULL;
        }
	if (!size_in) return NEW_STRING("");
	if (size_in < 0) {
		readall = TRUE;
		size = BUFSIZE(self);
	} else
		size = (size_t) size_in;
	list_obj = NEW_LIST(10);
	aprerr = 0;
	while (aprerr != APR_EOF && (readall || total_read < size)) {
		str_obj = NEW_OBJ(OBJ(STRING_PROTO));
		if (size < IMMEDIATE_DATA_LEN) {
			SET_TYPE_IF_EXC(OBJ(STRING_PROTO), str_obj, DATA_TYPE_IMMDATA) return NULL;
			buf = str_obj->data.str;
		} else {
			obj_str = obj_malloc(ist, OBJ(STRING_PROTO), str_obj, sizeof(pr_str_t)+size+1);
			buf = obj_str->str;
		}
		aprerr = apr_file_gets(buf, (int)(size - total_read + 1), STREAM(self));
		if (aprerr != APR_SUCCESS && aprerr != APR_EOF) {
			raise_exception(ist, OBJ(IOEXCEPTION), "%s",
					apr_strerror(aprerr,
					err_buf, sizeof(err_buf)));
			return NULL;
		}
		total_read += num_read = strlen(buf);
		while ( aprerr != APR_EOF && buf[num_read-1] != '\n' &&
				                   (readall || total_read < size) ) {
			size_t next_read = (readall ? size : size - total_read);
			size_t new_len = num_read + next_read;
			if (str_obj->data_type == DATA_TYPE_IMMDATA && new_len >= IMMEDIATE_DATA_LEN) {
				obj_str = obj_malloc(ist, OBJ(STRING_PROTO), str_obj, sizeof(pr_str_t)+new_len+1);
				buf = obj_str->str;
			} else {
				obj_str = obj_realloc(ist, str_obj, sizeof(pr_str_t)+new_len+1);
				buf = obj_str->str;
			}
			aprerr = apr_file_gets(buf + num_read, (int)(next_read + 1), STREAM(self));
			if (aprerr != APR_SUCCESS && aprerr != APR_EOF) {
				raise_exception(ist, OBJ(IOEXCEPTION), "%s",
						apr_strerror(aprerr,
						err_buf, sizeof(err_buf)));
				return NULL;
			}
			total_read += num_read = strlen(buf);
		}
		if (str_obj->data_type == DATA_TYPE_IMMDATA)
			str_obj->imm_data_len = (int) num_read;
		else {
			obj_str = obj_realloc(ist, str_obj, sizeof(pr_str_t)+num_read+1);
			buf = obj_str->str;
			obj_str->len = num_read;
		}
		buf[num_read] = 0;
		str_obj->immutable = TRUE;
		list_append(ist, list_obj, str_obj);
	}
	return list_obj;
}

DEF(File, seek, FPARM2( pos, NULL,
                        how, NEW_INT(0) ) ) {
	apr_off_t pos;
	apr_seek_where_t origin;
	apr_status_t aprerr;
	char err_buf[1024];
	i64_t pos_in, how;

	BIN_CONTENT_CHK(File);
	INT_64_PARAM(1, pos_in);
	INT_64_PARAM(2, how);

	if (!has_proto(ist, self, File_OBJ)) {
		raise_exception(ist, OBJ(TYPE_EXC), "seek can only be called on a File object");
		return NULL;
	}
	if (!OPEN(self)) {
		raise_exception(ist, OBJ(IOEXCEPTION), "seek called on closed file");
		return NULL;
	}
	pos = (apr_off_t)pos_in;
	switch(how) {
		case 0: origin = APR_SET; break;
		case 1: origin = APR_CUR; break;
		case 2: origin = APR_END; break;
		default :
			raise_exception(ist, OBJ(TYPE_EXC), "file seek how parameter must be 0, 1, or 2");
			return NULL;
	}

	aprerr = apr_file_seek(STREAM(self), origin, &pos);
	if (aprerr != APR_SUCCESS) {
		raise_exception(ist, OBJ(IOEXCEPTION), "%s",
				apr_strerror(aprerr, err_buf, sizeof(err_buf)));
		return NULL;
	}
	return OBJ(NONE);
}
#if 0
DEF(File, tell, NULL) {
	BIN_CONTENT_CHK(File);
	if (!OPEN(self)) {
		raise_exception(ist, OBJ(IOEXCEPTION), "tell called on closed file");
		return NULL;
	}
	return NEW_INT(ftell(STREAM(self)));
}
#endif
DEF(File, truncate, FPARM1(size, NULL)) {
	apr_status_t aprerr;
	char err_buf[1024];
	apr_off_t size;
	i64_t size_in;

	BIN_CONTENT_CHK(File);
	INT_64_PARAM(1, size_in);

	if (!has_proto(ist, self, File_OBJ)) {
		raise_exception(ist, OBJ(TYPE_EXC), "truncate can only be called on a File object");
		return NULL;
	}
	if (!OPEN(self)) {
		raise_exception(ist, OBJ(IOEXCEPTION), "truncate called on closed file");
		return NULL;
	}

	/* XXX Need a way to check for upper bounds portably. This can be
	 * different on 64-bit systems, or systems which support LFS */
	if (size_in < 0) {
		raise_exception(ist, OBJ(TYPE_EXC), "file truncate size value out of range");
		return NULL;
	}
	size = (apr_off_t)(size_in);

	aprerr = apr_file_trunc(STREAM(self), size);
	if(aprerr != APR_SUCCESS) {
		raise_exception(ist, OBJ(IOEXCEPTION), "%s",
				apr_strerror(aprerr, err_buf, sizeof(err_buf)));
		return NULL;
	}
	return OBJ(NONE);
}

DEF(File, write, FPARM1(str, NULL)) {
	apr_size_t size;
	apr_status_t aprerr;
	char err_buf[1024];
	char* str;

	BIN_CONTENT_CHK(File);
	STRING_PARAM(1, str);
	size = pr_strlen(parms[1]);

	if (!has_proto(ist, self, File_OBJ)) {
		raise_exception(ist, OBJ(TYPE_EXC), "write can only be called from a File object");
		return NULL;
	}
	if (!OPEN(self)) {
		raise_exception(ist, OBJ(IOEXCEPTION), "write called on closed file");
		return NULL;
	}
	if (!(FLAGS(self) & APR_WRITE)) {
		raise_exception(ist, OBJ(IOEXCEPTION), "write called on file opened without write perms");
		return NULL;
	}

	aprerr = apr_file_write(STREAM(self), str, &size);
	if (aprerr != APR_SUCCESS) {
		raise_exception(ist, OBJ(IOEXCEPTION), "%s",
				apr_strerror(aprerr, err_buf, sizeof(err_buf)));
		return NULL;
	}
	return parms[1];
}

DEF(File, writeLines, FPARM1(lst, NULL)) {
	char* str;
	apr_size_t size;
	apr_status_t aprerr;
	char err_buf[1024];

	if (!has_proto(ist, parms[1], OBJ(SEQ_PROTO)) && !has_proto(ist, parms[1], OBJ(STRING_PROTO))) {
		raise_exception(ist, OBJ(TYPE_EXC), "file writeLines parameter must be a sequence or a string");
		return NULL;
	}
	if (!has_proto(ist, self, File_OBJ)) {
		raise_exception(ist, OBJ(TYPE_EXC), "writeLines can only be called on a File object");
		return NULL;
	}
	if (!OPEN(self)) {
		raise_exception(ist, OBJ(IOEXCEPTION), "write called on closed file");
		return NULL;
	}
	if (!(FLAGS(self) & APR_WRITE)) {
		raise_exception(ist, OBJ(IOEXCEPTION), "writeLines called on file opened without write perms");
		return NULL;
	}
	if (has_proto(ist, parms[1], OBJ(STRING_PROTO))) {
		str  = pr_strptr(parms[1]);
		size = pr_strlen(parms[1]);
		aprerr = apr_file_write(STREAM(self), str, &size);
		if (aprerr != APR_SUCCESS) {
			raise_exception(ist, OBJ(IOEXCEPTION), "%s",
					apr_strerror(aprerr, err_buf, sizeof(err_buf)));
			return NULL;
		}
	} else {
		int i, llen = (int) list_len(ist, parms[1]);
		for(i = 0; i < llen; i++) {
			obj_p item = list_item(ist, parms[1], i);
			if (!has_proto(ist, item, OBJ(STRING_PROTO))) {
				raise_exception(ist, OBJ(TYPE_EXC), "file writeLines lst item must be a string");
				return NULL;
			}
			str  = pr_strptr(item);
			size = pr_strlen(item);
			aprerr = apr_file_write(STREAM(self), str, &size);
			if (aprerr != APR_SUCCESS) {
				raise_exception(ist, OBJ(IOEXCEPTION), "%s",
						apr_strerror(aprerr, err_buf, sizeof(err_buf)));
				return NULL;
			}
		}
	}
	return parms[1];
}

DEF(File, setStdOut, FORM_RPARAM) {
	int i;
	if (!has_proto(ist, parms[1], File_OBJ)) {
		raise_exception(ist, OBJ(TYPE_EXC), "stdout can only be assigned to a file object");
		return NULL;
	}
	su(i);  read_unlock(ist, self);
	set_attr(ist, self, sym(ist, "stdout"), parms[1]);
	un_su(i); read_lock(ist, self);
	return parms[1];
}


#define __STDIN		0
#define __STDOUT	1
#define __STDERR	2

static obj_p __open_std_file(int type, apr_pool_t *subpool)
{
	obj_p mode, file;
	apr_file_t *fp;
	apr_status_t aprerr;
	apr_int32_t flags;

	switch (type) {
		case __STDIN:
			aprerr = apr_file_open_stdin(&fp, subpool);
			mode = NEW_STRING("r");
			file = NEW_STRING("<stdin>");
			flags = APR_READ;
			break;

		case __STDOUT:
			aprerr = apr_file_open_stdout(&fp, subpool);
			file = NEW_STRING("<stdout>");
			mode = NEW_STRING("w");
			flags = APR_WRITE;
			break;

		case __STDERR:
			aprerr = apr_file_open_stderr(&fp, subpool);
			file = NEW_STRING("<stderr>");
			mode = NEW_STRING("w");
			flags = APR_WRITE;
			break;

		default:
			return NULL; /* XXX something stupid, do an error */
	}

	if (aprerr != APR_SUCCESS)
		return NULL;

	return new_file_object(file, mode, fp, MIN_BUFSIZE, flags);
}

MAIN_MODULE_INIT(File) {
	apr_pool_t *std_pool;
	apr_status_t aprerr;
	obj_p file_obj;

	MODULE_SUB_INIT(File);
	MODULE_ADD_SYM(File, init_);
	MODULE_ADD_SYM(File, read);
	MODULE_ADD_SYM(File, readLine);
	MODULE_ADD_SYM(File, readLines);
	MODULE_ADD_SYM(File, write);
	MODULE_ADD_SYM(File, writeLines);
	MODULE_ADD_SYM(File, flush);
	MODULE_ADD_SYM(File, seek);
	MODULE_ADD_SYM(File, truncate);
	MODULE_ADD_SYM(File, close);
	MODULE_ADD_SYM(File, closed_QUES);
	MODULE_ADD_SYM(File, setStdOut);
	//MODULE_ADD_SYM(File, fileNo);
	//MODULE_ADD_SYM(File, isATty_QUES);
	//MODULE_ADD_SYM(File, tell);
	MODULE_ADD_SYM(File, objList_);

	/* Create base file objects for std{in,out,err} */
	aprerr = apr_pool_create(&std_pool, get_pr_head_pool());
	/* Need some kind of error return! */
	if (aprerr != APR_SUCCESS)
		return;

	if ((file_obj = __open_std_file(__STDIN, std_pool)))
		set_attr(ist, File_OBJ, sym(ist, "stdin"), file_obj);
	if ((file_obj = __open_std_file(__STDOUT, std_pool)))
		set_attr(ist, File_OBJ, sym(ist, "stdout"), file_obj);
	if ((file_obj = __open_std_file(__STDERR, std_pool)))
		set_attr(ist, File_OBJ, sym(ist, "stderr"), file_obj);

	check_exceptions(ist);
}
