/* ====================================================================
 * The Prothon License Agreement, Version 1.1
 *
 * Copyright (c) 2004 Hahn Creative Applications, http://hahnca.com.
 * All rights reserved. 
 *
 * 1. This LICENSE AGREEMENT is between Hahn Creative Applications ("HCA"),
 * and the Individual or Organization ("Licensee") accessing and otherwise
 * using Prothon software in source or binary form and its associated
 * documentation.
 * 
 * 2. Subject to the terms and conditions of this License Agreement, HCA
 * hereby grants Licensee a nonexclusive, royalty-free, world-wide license
 * to reproduce, analyze, test, perform and/or display publicly, prepare
 * derivative works, distribute, and otherwise use Prothon alone or in any
 * derivative version, provided, however, that HCA's License Agreement and
 * HCA's notice of copyright, i.e., "Copyright (c) 2004 Hahn Creative
 * Applications; All Rights Reserved" are retained in Prothon alone or
 * in any derivative version prepared by Licensee.
 * 
 * 3. In the event Licensee prepares a derivative work that is based on or
 * incorporates Prothon or any part thereof, and wants to make the
 * derivative work available to others as provided herein, then Licensee
 * hereby agrees to include in any such work a brief summary of the
 * changes made to Prothon.
 * 
 * 4. HCA is making Prothon available to Licensee on an "AS IS" basis.
 * HCA MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.  BY WAY
 * OF EXAMPLE, BUT NOT LIMITATION, HCA MAKES NO AND DISCLAIMS ANY
 * REPRESENTATION OR WARRANTY OF MERCHANTABILITY OR FITNESS FOR ANY
 * PARTICULAR PURPOSE OR THAT THE USE OF PROTHON WILL NOT INFRINGE ANY
 * THIRD PARTY RIGHTS.
 * 
 * 5. HCA SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PROTHON
 * FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL DAMAGES OR LOSS AS A
 * RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PROTHON, OR ANY
 * DERIVATIVE THEREOF, EVEN IF ADVISED OF THE POSSIBILITY THEREOF.
 * 
 * 6. This License Agreement will automatically terminate upon a material
 * breach of its terms and conditions.
 * 
 * 7. Nothing in this License Agreement shall be deemed to create any
 * relationship of agency, partnership, or joint venture between HCA and
 * Licensee.  This License Agreement does not grant permission to use HCA
 * trademarks or trade name in a trademark sense to endorse or promote
 * products or services of Licensee, or any third party.
 * 
 * 8. By copying, installing or otherwise using Prothon, Licensee agrees
 * to be bound by the terms and conditions of this License Agreement.
 * ====================================================================
 */


// builtins.c

#include <ctype.h>
#include <stdio.h>
#include <string.h>

#include <apr_strings.h>

#include <prothon/prothon.h>
#include "dict.h"
#include "parser.h"
#include "object.h"
#include <prothon/prothon_dll.h>

#include "pr_snprintf.h"

#define is_String(objid)        (has_proto(ist, objid, String_OBJ))

MODULE_DECLARE(String);
MODULE_DECLARE(StringGen);

//********************************* new_string_obj ****************************
// this cannot be used with binary data, for C strings only
// for binary data with embedded nulls use NEW_STRINGN()
obj_p new_string_obj(isp ist, const char* str){
	obj_p obj;
	pr_str_p obj_str;
	size_t len;
	if(!str) return NULL;
	len = strlen(str);
	obj = NEW_OBJ(OBJ(STRING_PROTO));
	if (len < IMMEDIATE_DATA_LEN) {
		SET_TYPE_IF_EXC(OBJ(STRING_PROTO), obj, DATA_TYPE_IMMDATA) return NULL;
		obj->imm_data_len = (int) len;
		strcpy(obj->data.str, str);
	} else {
		obj_str = obj_malloc(ist, OBJ(STRING_PROTO), obj, sizeof(pr_str_t)+len+1); if_exc_return NULL;
		obj_str->len = len;
		strcpy(&(obj_str->str[0]), str);
	}
	obj->immutable = TRUE;
	return obj;
}

//********************************* new_string_n_obj **************************
obj_p new_string_n_obj(isp ist, char* str, size_t len) {
	obj_p obj;
	pr_str_p obj_str;
	if(!str) return NULL;
	obj = NEW_OBJ(OBJ(STRING_PROTO));
	if (len < IMMEDIATE_DATA_LEN) {
		SET_TYPE_IF_EXC(OBJ(STRING_PROTO), obj, DATA_TYPE_IMMDATA) return NULL;
		obj->imm_data_len = (int) len;
		memcpy(obj->data.str, str, len);
		obj->data.str[len] = 0;
	} else {
		obj_str = obj_malloc(ist, OBJ(STRING_PROTO), obj, sizeof(pr_str_t)+len+1);
		obj_str->len = len;
		memcpy(&(obj_str->str[0]), str, len);
		obj_str->str[len] = 0;
	}
	obj->immutable = TRUE;
	return obj;
}

//********************************* set_string_data ***************************
void set_string_data(isp ist, obj_p obj, char* p, size_t len) {
	pr_str_p obj_str;
	if (len < 8) {
		SET_TYPE_IF_EXC(OBJ(STRING_PROTO), obj, DATA_TYPE_IMMDATA) return;
		obj->imm_data_len = (u8_t) len;
		memcpy(obj->data.str, p, len);
		obj->data.str[len]  = 0;
	} else {
		obj_str = obj->data.ptr = obj_malloc(ist, OBJ(STRING_PROTO), obj, sizeof(pr_str_t)+len+1);
		obj_str->len = len;
		memcpy(&(obj_str->str[0]), p, len);
		obj_str->str[len] = 0;
	}
	obj->immutable = TRUE;
}

//********************************* changeCase ********************************
// Generic case-changer function.
#define AT_WORD_START -1
static obj_p changeCase(isp ist,obj_p self,int (*changer)(int), size_t howMany) {
	size_t str_len = pr_strlen(self), ii = 0;
	char *src_ptr, *dest_ptr = 0;
	obj_p obj;
	int atStart = 1;

	obj = copy_object(ist, self);
	dest_ptr  = pr_strptr(obj);

	// Change the copy.
	src_ptr = pr_strptr(self);
	for (ii=0; ii<str_len; ++ii) {
		if (howMany == AT_WORD_START) {
			if (atStart)
				dest_ptr[ii] = (char)(*changer)(src_ptr[ii]);
			if (isspace(src_ptr[ii])) atStart = 1; else atStart = 0;
		} else {
			if (ii<howMany)
				dest_ptr[ii] = (char)(*changer)(src_ptr[ii]);
		}
	}
	return obj;
}

//********************************* bin_strstr ********************************
// strstr()-analog that works for strings that may contain nulls.
// Totally dependent upon ASCII encoding; beware.
static char* bin_strstr(char* haystack,size_t lhaystack, char* needle,size_t lneedle) {
	char* lookfor = needle;
	char* looking_at = haystack;
	char* hs_end = haystack+lhaystack;
	char* ndl_end = needle+lneedle;
	while (looking_at < hs_end) {
		if (*lookfor == *looking_at) {
			// Matched a character.
			++lookfor;
			if (lookfor == ndl_end) {
				// Matched the whole string. It started lneedle-1
				// characters before the character we're looking at
				// now.
				return looking_at + 1 - lneedle;
			}
		} else {
			// No match.
			if (lookfor > needle) {
				// We had recognized the beginning of a match. We must
				// return to the character after the location at which
				// the non-match began.
				looking_at -= lookfor-needle;
				lookfor = needle;
			}
		}
		++looking_at;
	}
	return 0;
}

//********************************* STRING MODULE *****************************

MODULE_START(String)
{
	String_OBJ = OBJ(STRING_PROTO);
	MODULE_SET_DOC(String, "string object prototype");
	set_obj_id(String_OBJ, *, String);
	set_attr(ist, OBJ(OBJECT), sym(ist, "String"), String_OBJ);
	SET_TYPE_IF_EXC(String_OBJ, String_OBJ, DATA_TYPE_IMMDATA) return;
	String_OBJ->imm_data_len = 0;
	String_OBJ->data.str[0]  = 0;
}

DEF(String, init_, FORM_STAR_PARAM) {
	char *s;
	obj_p res = NULL;
	int llen;
	BIN_EMPTY_CHK();
	llen = (int) list_len(ist, parms[1]);
	if (llen > 2) {
		raise_exception(ist, OBJ(TYPE_EXC), "incorrect number of parameters");
		return NULL;
	}
	if (llen > 0) {
		obj_p obj2, obj = list_item(ist, parms[1], 0);
		if (llen == 1 && has_proto(ist, obj, OBJ(SYMBOL_PROTO))) {
			s = symch(ist, obj);
		} else if ( llen == 2 && has_proto(ist,      obj, OBJ(INT_PROTO)) && 
			        has_proto(ist, obj2 = list_item(ist, parms[1], 1), OBJ(INT_PROTO)) ) {
			if (obj->data_type == DATA_TYPE_IMMDATA) {
				res = long_format(new_longp(obj->data.i64), int2i32t(ist, obj2), FALSE);
				s = pr_strptr(res);
			} else {
				res = long_format(obj->data.ptr, int2i32t(ist, obj2), FALSE);
				s = pr_strptr(res);
			}
		} else if (llen == 1) {
			s = as_str(ist, obj);
		} else {
			raise_exception(ist, OBJ(TYPE_EXC), "invalid parameters");
			return NULL;
		}
	} else
		s = "";
	set_string_data(ist, self, s, strlen(s));
	if (res) del_unlock(res);
	return OBJ(NONE);
}

DEF(String, str_, NULL) {
	BIN_STR_CHK(String);
	return self;
}

DEF(String, hash_, NULL) {
	i32_t res = 0x9a7c5e3;
	char *p;
	BIN_CONTENT_CHK(String);
	for(p=pr_strptr(self); *p; p++) res += 131 * (*p);
	return new_hash_obj(ist, res);
}

DEF(String, getItem_, FORM_RPARAM) {
	BIN_CONTENT_CHK(String);
	return get_sequence_item(ist, self, parms[1], SEQ_TYPE_STRING);
}

DEF(String, setItem_, FORM_RPARAM) {
	raise_exception(ist, OBJ(MUTABLE_EXC), "strings may not be modifed");
	return NULL;
}

DEF(String, delItem_, NULL) {
	raise_exception(ist, OBJ(MUTABLE_EXC), "strings may not be modifed");
	return NULL;
}

DEF(String, cmp, FORM_RPARAM){
	obj_p other = parms[1];
	BIN_CONTENT_CHK(String);
	if (self == other)
		return NEW_INT(0);
	if (!is_String(other)) {
		raise_exception(ist, OBJ(TYPE_EXC), "Cannot compare a string and non-string");
		return NULL;
	}
	return NEW_INT(strcmp(pr_strptr(self), pr_strptr(other)));
}

DEF(String, add_, FORM_RPARAM){
	obj_p obj, other = parms[1];
	pr_str_p obj_str;
	size_t slen, olen, tlen;
	BIN_CONTENT_CHK(String);
	if (!is_String(other)) other = call_func0(ist, other, SYM(STR_));
	slen = pr_strlen(self); olen = pr_strlen(other);
	tlen = slen + olen;
	obj = NEW_OBJ(String_OBJ);
	if (tlen < IMMEDIATE_DATA_LEN) {
		SET_TYPE_IF_EXC(String_OBJ, obj, DATA_TYPE_IMMDATA) return NULL;
		obj->imm_data_len = (int) tlen;
		memcpy(&(obj->data.str[0]),    pr_strptr(self),  slen);
		memcpy(&(obj->data.str[slen]), pr_strptr(other), olen);
		obj->data.str[tlen] = 0;
	} else {
		obj_str = obj_malloc(ist, OBJ(STRING_PROTO), obj, sizeof(pr_str_t)+tlen+1);
		obj_str->len = tlen;
		memcpy(&(obj_str->str[0]),    pr_strptr(self),  slen);
		memcpy(&(obj_str->str[slen]), pr_strptr(other), olen);
		obj_str->str[tlen] = 0;
	}
	obj->immutable = TRUE;
	return obj;
}

DEF(String, mul_, FORM_RPARAM){
	obj_p obj;
	pr_str_p obj_str;
	size_t i, times, tlen, len = pr_strlen(self);
	BIN_CONTENT_CHK(String);
	if (!has_proto(ist, parms[1], OBJ(INT_PROTO))) {
		raise_exception(ist, OBJ(TYPE_EXC), "multiply times parameter must be an integer");
		return NULL;
	}
	times = (size_t) parms[1]->data.i64;
	if(times == 0) return NEW_STRING("");
	if(times == 1) return self;
	tlen = len*times;
	obj = NEW_OBJ(String_OBJ);
	if (tlen < IMMEDIATE_DATA_LEN-1) {
		SET_TYPE_IF_EXC(String_OBJ, obj, DATA_TYPE_IMMDATA) return NULL;
		obj->imm_data_len = (int) tlen;
		for(i=0; i < times; i++)
			memcpy(obj->data.str+i*len, pr_strptr(self), len);
		obj->data.str[tlen] = 0;
	} else {
		obj_str = obj_malloc(ist, String_OBJ, obj, sizeof(pr_str_t)+tlen+1);
		if(!obj_str) {
			raise_exception(ist, OBJ(OUTOFMEMORY_EXC), "memory allocation failed for string multiplication");
			return NULL;
		}
		obj_str->len = tlen;
		for(i=0; i < times; i++)
			memcpy(obj_str->str+i*len, pr_strptr(self), len);
		obj_str->str[tlen] = 0;
	}
	obj->immutable = TRUE;
	return obj;
}


DEF(String, mod_, FORM_RPARAM) {
	obj_p			res, argv_obj = parms[1];
	size_t			i, argc;
	pr_arg_p		argv;
	char			*fmt_str = pr_strptr(self), *buf;
	apr_status_t	aprerr;

	BIN_CONTENT_CHK(String);
	if ( has_proto(ist, argv_obj, OBJ(STRING_PROTO)) ||
		!has_proto(ist, argv_obj,    OBJ(SEQ_PROTO)) ) {
		argv_obj = NEW_LIST(1);
		list_append(ist, argv_obj, parms[1]);
	}
	argc = list_len(ist, argv_obj);
	argv = pr_malloc((argc+1) * sizeof(pr_arg_t));
	for (i=0; i < argc; i++) {
		obj_p arg_obj = list_item(ist, argv_obj, (int) i);
		if (has_proto(ist, arg_obj, OBJ(INT_PROTO))) {
			argv[i].arg.i64 = arg_obj->data.i64;
			argv[i].type    = ARG_TYPE_INT;
		} else if (has_proto(ist, arg_obj, OBJ(FLOAT_PROTO))) {
			argv[i].arg.f64 = arg_obj->data.f64;
			argv[i].type    = ARG_TYPE_FLOAT;
		} else if (has_proto(ist, arg_obj, OBJ(STRING_PROTO))) {
			argv[i].arg.str = pr_strptr(arg_obj);
			argv[i].type    = ARG_TYPE_STRING;
		} else {
			raise_exception(ist, OBJ(TYPE_EXC), "invalid type for string (%%) format");
			return NULL;
		}
	}
	argv[i].type = ARG_TYPE_END;
	buf = pr_malloc(16384);
	aprerr = apr_pr_vsnprintf(buf, 16384, fmt_str, argv);
	if (aprerr == -99) {
		raise_exception(ist, OBJ(INTERPRETER_EXC), "invalid or wrong number of %% params");
		return NULL;
	}
	res = NEW_STRING(buf);
	pr_free(buf);
	return res;
}

DEF(String, ord, NULL) {
	BIN_CONTENT_CHK(String);
	return NEW_INT((int)(*pr_strptr(self)));
}

DEF(String, join, FORM_RPARAM) {
	char *self_str, *dest_ptr;
	pr_str_p obj_str;
	obj_p obj, list, str_list;
	int i, llen;
	size_t ofs, self_len, tlen;
	BIN_CONTENT_CHK(String);
	if (!has_proto(ist, parms[1], OBJ(LIST_PROTO))) {
		raise_exception(ist, OBJ(TYPE_EXC), "join function parameter must be a list");
		return NULL;
	}
	list = parms[1];
	llen = (int) list_len(ist, parms[1]);
	if (!llen) return NEW_STRING("");
	str_list = NEW_LIST(llen);
	self_str = pr_strptr(self);
	self_len = pr_strlen(self);
	tlen = 0;
	for (i=0; i < llen; i++) {
		obj_p str_obj = call_func0(ist, list_item(ist, list, i), SYM(STR_)); if_exc_return NULL;
		list_append(ist, str_list, str_obj);
		tlen += pr_strlen(str_obj); 
		if (i != llen-1) tlen += self_len;
	}
	obj = NEW_OBJ(String_OBJ);
	if (tlen < IMMEDIATE_DATA_LEN) {
		SET_TYPE_IF_EXC(String_OBJ, obj, DATA_TYPE_IMMDATA) return NULL;
		obj->imm_data_len = (int) tlen;
		dest_ptr = obj->data.str;
	} else {
		obj_str = obj_malloc(ist, OBJ(STRING_PROTO), obj, sizeof(pr_str_t)+tlen+1);
		if(!obj_str) {
			raise_exception(ist, OBJ(OUTOFMEMORY_EXC), "memory allocation failed for string join");
			return NULL;
		}
		obj_str->len = tlen;
		dest_ptr = obj_str->str;
	}
	for(i=0, ofs=0; i < llen; i++) {
		obj_p  str_obj = list_item(ist, str_list, i);
		size_t str_len = pr_strlen(str_obj);
		memcpy(dest_ptr+ofs, pr_strptr(str_obj), str_len);
		ofs += str_len;
		if (i != llen-1) {
			memcpy(dest_ptr+ofs, self_str, self_len);
			ofs += self_len;
		}
	}
	dest_ptr[tlen] = 0;
	set_immutable(obj);
	del_unlock(str_list);
	return obj;
}

DEF(String, in__QUES, FORM_RPARAM) {
	BIN_CONTENT_CHK(String);
	if (!has_proto(ist, parms[1], String_OBJ))
		return call_func1(ist, parms[1], SYM(RIN__QUES), self);
	if (strstr(pr_strptr(parms[1]), pr_strptr(self))) return OBJ(PR_TRUE);
	else						              return OBJ(PR_FALSE);
}

DEF(String, notIn__QUES, FORM_RPARAM) {
	BIN_CONTENT_CHK(String);
	if (!has_proto(ist, parms[1], String_OBJ))
		return call_func1(ist, parms[1], SYM(RNOTIN__QUES), self);
	if (strstr(pr_strptr(parms[1]), pr_strptr(self))) return OBJ(PR_FALSE);
	else						              return OBJ(PR_TRUE);
}

DEF(String, len, NULL) {
	BIN_CONTENT_CHK(String);
	return NEW_INT(pr_strlen(self));
}

DEF(String, bool__QUES, NULL) {
	BIN_CONTENT_CHK(String);
	if (pr_strlen(self)) return OBJ(PR_TRUE);
	else                 return OBJ(PR_FALSE);
}

DEF(String, lower, NULL) {
	BIN_CONTENT_CHK(String);
	return changeCase(ist,self,&tolower,pr_strlen(self));
}

DEF(String, upper, NULL) {
	BIN_CONTENT_CHK(String);
	return changeCase(ist,self,&toupper,pr_strlen(self));
}

DEF(String, capitalize, NULL) {
	BIN_CONTENT_CHK(String);
	return changeCase(ist,self,&toupper,1);
}

DEF(String, capWords, NULL) {
	BIN_CONTENT_CHK(String);
	return changeCase(ist,self,&toupper,AT_WORD_START);
}

#define int_imm_value(objid)	(objid->data.i64)
DEF(String, find, FPARM2( stringToFind, NULL, indexToStartLooking, NEW_INT(0) )) {
	size_t str_len = pr_strlen(self);
	size_t find_len = 0;
	char *found, *find_ptr = 0;
	size_t start = 0;
	obj_p result = 0;

	BIN_CONTENT_CHK(String);

	if (!has_proto(ist, parms[1], String_OBJ)) {
		raise_exception(ist, OBJ(TYPE_EXC), "find function parameter 1 must be a string");
		return NULL;
	}
	if (!has_proto(ist, parms[3], OBJ(INT_PROTO))) {
		raise_exception(ist, OBJ(TYPE_EXC), "find function parameter 2 must be an integer");
		return NULL;
	} else {
		start = (size_t)int_imm_value(parms[3]);
	}

	find_ptr = pr_strptr(parms[1]);
	find_len = pr_strlen(parms[1]);
	if (find_len<1) {
		return new_int_obj(ist,0);
	}
	found = bin_strstr(pr_strptr(self)+int_imm_value(parms[3]),str_len,find_ptr,find_len);
	if (found) {
		result = new_int_obj(ist,found-pr_strptr(self));
	} else {
		result = new_int_obj(ist,0);
	}

	return result;
}

DEF(String, replace, FORM_PARAM2) {
	size_t str_len = pr_strlen(self);
	size_t find_len = 0;
	size_t repl_len = 0;
	size_t result_len = str_len;
	char* dest_ptr = 0;
	obj_p obj=0;
	pr_str_p obj_str = 0;
	char* src_ptr = 0;
	char* find_ptr = 0;
	char* repl_ptr = 0;
	size_t temp_len = 0;
	char* a_match = 0;
	char* where_ptr = 0;

	BIN_CONTENT_CHK(String);

	if (!has_proto(ist, parms[1], String_OBJ)) {
		raise_exception(ist, OBJ(TYPE_EXC), "replace function parameter 1 must be a string");
		return NULL;
	}
	if (!has_proto(ist, parms[3], String_OBJ)) {
		raise_exception(ist, OBJ(TYPE_EXC), "replace function parameter 2 must be a string");
		return NULL;
	}

	find_len = pr_strlen(parms[1]);
	if (find_len>str_len) return self;
	if (find_len<1) {
		raise_exception(ist, OBJ(TYPE_EXC), "replace: empty search string");
		return NULL;
	}
	repl_len = pr_strlen(parms[3]);
	src_ptr = pr_strptr(self);
	find_ptr = pr_strptr(parms[1]);
	repl_ptr = pr_strptr(parms[3]);

	// This is perhaps not as efficient as it could be.

	// Find the length of the result string.
	temp_len = str_len;
	a_match = bin_strstr(src_ptr,temp_len,find_ptr,find_len);
	while (a_match) {
		result_len -= find_len;
		result_len += repl_len;
		a_match += find_len; // Skip the matching text.
		temp_len = str_len - (a_match - src_ptr);
		a_match = bin_strstr(a_match,temp_len,find_ptr,find_len);
	}

	// Create the copy.
	// Create the copy.
	obj = NEW_OBJ(String_OBJ);
	if (result_len < IMMEDIATE_DATA_LEN-1) {
		SET_TYPE_IF_EXC(String_OBJ, obj, DATA_TYPE_IMMDATA) return NULL;
		obj->imm_data_len = (int) result_len;
		dest_ptr = obj->data.str;
		dest_ptr[result_len]=0;
	} else {
		obj_str = obj_malloc(ist, OBJ(STRING_PROTO), obj, sizeof(pr_str_t)+result_len+1);
		if(!obj_str) {
			raise_exception(ist, OBJ(OUTOFMEMORY_EXC), "memory allocation failed for string.replace");
			return NULL;
		}
		obj_str->len = result_len;
		dest_ptr = obj_str->str;
		dest_ptr[result_len]=0;
	}

	// Fill the replaced copy.
	temp_len = str_len;
	where_ptr = src_ptr;
	a_match = bin_strstr(where_ptr,temp_len,find_ptr,find_len);
	while (a_match) {
		// Copy everything from where we were to the beginning of the
		// match.
		int frag_len = (int) (a_match - where_ptr);
		memcpy(dest_ptr,where_ptr,frag_len);
		dest_ptr += frag_len;

		// Copy the replacement text.
		memcpy(dest_ptr,repl_ptr,repl_len);
		dest_ptr += repl_len;

		a_match += find_len; // Skip the matched text.
		where_ptr = a_match;
		temp_len = str_len - (a_match - where_ptr);
		a_match = bin_strstr(a_match,temp_len,find_ptr,find_len);
	}

	// Copy any trailing text after the last match.
	if (where_ptr<src_ptr+str_len) {
		memcpy(dest_ptr,where_ptr,(src_ptr+str_len)-where_ptr);
	}
	return obj;
}

DEF(String, expandTabs, FPARM1( nSpaces, NEW_INT(8) )) {
	obj_p result;	
	BIN_CONTENT_CHK(String);
	{
		obj_p tab = new_string_obj(ist,"\t");
		obj_p tspace = new_string_obj(ist," ");
		obj_p mul_args[]= {NULL, parms[1]};
		obj_p spcs = Stringmul_(ist,tspace,2,mul_args,0);
		obj_p repl_args[] = {NULL, tab, NULL, spcs};
		result = Stringreplace(ist,self,4,repl_args,0);
		del_unlock(tab);
		del_unlock(spcs);
		del_unlock(tspace);
	}
	return result;
}

//===============================
// ASSUMPTION: if we're using lStrip, rStrip, etc, then
// we're dealing with null-terminated text strings.
//===============================
DEF(String, lStrip, NULL) {
	char* cstr = pr_strptr(self);

	BIN_CONTENT_CHK(String);

	while (isspace(*cstr)) {
		++cstr;
	}
	
	// Create the copy.
	return new_string_obj(ist,cstr);
}

DEF(String, rStrip, NULL) {
	// Count whitespace at the end.
	char* cstr = pr_strptr(self)+pr_strlen(self)-1;
	size_t result_len = 0;

	BIN_CONTENT_CHK(String);

	while (isspace(*cstr)) {
		--cstr;
	}
	result_len = cstr-pr_strptr(self)+1;
	
	// Create the copy.
	return new_string_n_obj(ist,pr_strptr(self),result_len);
}

DEF(String, strip, NULL) {
	char *cstr2, *cstr = pr_strptr(self);
	size_t result_len;

	BIN_CONTENT_CHK(String);

	while (isspace(*cstr)) {
		++cstr;
	}
	
	// Count whitespace at the end.
	cstr2 = pr_strptr(self)+pr_strlen(self)-1;
	result_len = 0;
	while (isspace(*cstr2)) {
		--cstr2;
	}
	result_len = cstr2-cstr+1;
	
	// Create the copy.
	return new_string_n_obj(ist,cstr,result_len);
}

DEF(String, split, FORM_RPARAM) {
	size_t str_len = pr_strlen(self);
	size_t find_len = 0;
	char* src_ptr = 0;
	char* find_ptr = 0;
	size_t temp_len = 0;
	char* a_match = 0;
	char* where_ptr = 0;
	obj_p next_frag, list_obj = new_list_obj(ist,0);

	BIN_CONTENT_CHK(String);
	if (!has_proto(ist, parms[1], String_OBJ)) {
		raise_exception(ist, OBJ(TYPE_EXC), "split function parameter 1 must be a string");
		return NULL;
	}

	find_len = pr_strlen(parms[1]);
	if (find_len<1) {
		raise_exception(ist, OBJ(TYPE_EXC), "split: empty separator");
		return NULL;
	}
	src_ptr = pr_strptr(self);
	where_ptr = src_ptr;
	find_ptr = pr_strptr(parms[1]);

	temp_len = str_len;
	a_match = bin_strstr(src_ptr,temp_len,find_ptr,find_len);
	while (a_match) {
		// Make new string from where_ptr to a_match.
		obj_p next_frag = new_string_n_obj(ist,where_ptr,a_match-where_ptr);
		list_append(ist,list_obj,next_frag);

		a_match += find_len; // Skip the matching text.
		where_ptr = a_match;
 		temp_len = str_len - (a_match - src_ptr);
		a_match = bin_strstr(a_match,temp_len,find_ptr,find_len);
	}

	// Copy any trailing text after the last match.
	next_frag = new_string_n_obj(ist,where_ptr,str_len-(where_ptr-src_ptr));
	list_append(ist,list_obj,next_frag);
	
	return list_obj;
}

DEF(String, iter_, NULL) {
	obj_p gen_obj;
	BIN_CONTENT_CHK(String);
	gen_obj = NEW_OBJ(StringGen_OBJ);
	SET_TYPE_IF_EXC(StringGen_OBJ, gen_obj, DATA_TYPE_DATAPTR) return NULL;
	gen_obj->data.ptr = pr_strptr(self);
	set_attr(ist, gen_obj, sym(ist, "savedString"), self);
	return gen_obj;
}


DEF(String, cDataLen_, NULL) {
	BIN_CONTENT_CHK(String);
	return NEW_INT(sizeof(pr_str_t) + pr_strlen(self) + 1);
}

DEF(String, objList_, FORM_RPARAM) {
	return parms[1];
}

//********************************* STRINGGEN MODULE **************************

MODULE_START(StringGen)
{
	StringGen_OBJ = NEW_OBJ(NULL);
	set_attr(ist, OBJ(OBJECT), sym(ist, "StringGen"), StringGen_OBJ);
	set_obj_id(StringGen_OBJ, *, StringGen);
}

DEF(StringGen, next, NULL) {
        int ch;
        obj_p res = NULL;
        char* res_str;
		BIN_CONTENT_CHK(StringGen);
        def_write_lock(self);
        if ((ch = *((char*)self->data.ptr))) {
				self->data.ptr = (i64_t *)(((char*)self->data.ptr) + 1);
                res = NEW_OBJ(String_OBJ);
                res_str = (char*)obj_malloc(ist, OBJ(STRING_PROTO), res, 2);
                res_str[0] = ch;
                res_str[1] = 0;
        } else
                raise_exception(ist, OBJ(STOP_ITERATION_EXC), NULL);
        def_write_unlock(self);
        return res;
}

DEF(StringGen, objList_, FORM_RPARAM) {
	return parms[1];
}

//********************************* MODULES INIT ******************************

MAIN_MODULE_INIT(String)
{
	MODULE_SUB_INIT(String);
	MODULE_ADD_SYM(String, init_);
	MODULE_ADD_SYM(String, bool__QUES);
	MODULE_ADD_SYM(String, str_);
	MODULE_ADD_SYM(String, hash_);
	MODULE_ADD_SYM(String, cmp);
	MODULE_ADD_SYM(String, ord);
	MODULE_ADD_SYM(String, len);
	MODULE_ADD_SYM(String, getItem_);
	MODULE_ADD_SYM(String, add_);
	MODULE_ADD_SYM(String, mul_);
	MODULE_ADD_SYM(String, mod_);
	MODULE_ADD_SYM(String, iter_);
	MODULE_ADD_SYM(String, join);
	MODULE_ADD_SYM(String, in__QUES);
	MODULE_ADD_SYM(String, notIn__QUES);
	MODULE_ADD_SYM(String, lower);
	MODULE_ADD_SYM(String, upper);
	MODULE_ADD_SYM(String, replace);
	MODULE_ADD_SYM(String, capitalize);
	MODULE_ADD_SYM(String, capWords);
	MODULE_ADD_SYM(String, expandTabs);
	MODULE_ADD_SYM(String, lStrip);
	MODULE_ADD_SYM(String, rStrip);
	MODULE_ADD_SYM(String, strip);
	MODULE_ADD_SYM(String, split);
	MODULE_ADD_SYM(String, find);
	MODULE_ADD_SYM(String, setItem_);
	MODULE_ADD_SYM(String, delItem_);
	MODULE_ADD_SYM(String, objList_);
	MODULE_ADD_SYM(String, cDataLen_);

	MODULE_SUB_INIT(StringGen);
	MODULE_ADD_SYM(StringGen, next);
	MODULE_ADD_SYM(StringGen, objList_);
	
	set_immutable(String_OBJ);

	check_exceptions(ist);
}
